@@ -1,3 +1,4 @@
+import collections
 import datetime
 import inspect
 
@@ -196,6 +197,7 @@ VALUE_DOWNSAMPLERS = {
     'sum_squares': 'q', # sum of squares of all datapoints
     'std_dev': 'd', # standard deviation of all datapoints
     'count': 'c', # number of all datapoints
+    'median': 'n', # middle-ranked value of all datapoints
     'most_often': 'o', # the most often occurring value of all datapoints (key mnemonic: o for often)
     'least_often': 'r', # the least often occurring value of all datapoints (key mnemonic: r for rare)
     'frequencies': 'f', # for each value number of occurrences in all datapoints
@@ -223,6 +225,7 @@ DERIVE_OPERATORS = {
 VALUE_TYPES = {
     'numeric',
     'graph',
+    'nominal',
 }
 
 
@@ -243,6 +246,10 @@ class ResultsBase(object):
     def __getitem__(self, key):
         raise NotImplementedError
 
+    def _get_backend_cursor(self):
+        # Very internal. Just for debugging and testing.
+        raise NotImplementedError
+
 
 class Streams(ResultsBase):
     pass
@@ -269,6 +276,26 @@ class Datastream(object):
 
         self.backend = backend
 
+    def _switch_database(self, database_name):
+        # Very internal. Just for debugging and testing.
+        self.backend._switch_database(database_name)
+
+    def _check_query_tags(self, tags):
+        for key, value in tags.iteritems():
+            # One should use nested dicts and not `__`.
+            if '__' in key:
+                raise exceptions.ReservedTagNameError("Reserved tag name used: %s" % key)
+
+            if isinstance(value, collections.Mapping):
+                self._check_query_tags(value)
+
+    def _check_tags(self, tags):
+        for key, value in tags.iteritems():
+            if key in RESERVED_TAGS:
+                raise exceptions.ReservedTagNameError("Reserved tag name used: %s" % key)
+
+        self._check_query_tags(tags)
+
     def ensure_stream(self, query_tags, tags, value_downsamplers, highest_granularity, derive_from=None, derive_op=None, derive_args=None, value_type=None):
         """
         Ensures that a specified stream exists.
@@ -286,6 +313,9 @@ class Datastream(object):
         :return: A stream identifier
         """
 
+        self._check_query_tags(query_tags or {})
+        self._check_tags(tags or {})
+
         if highest_granularity not in Granularity.values:
             raise exceptions.UnsupportedGranularity("'highest_granularity' is not a valid value: '%s'" % highest_granularity)
 
@@ -328,9 +358,7 @@ class Datastream(object):
         :param tags: A dictionary of new tags
         """
 
-        for key in tags.keys():
-            if key in RESERVED_TAGS:
-                raise exceptions.ReservedTagNameError
+        self._check_tags(tags)
 
         self.backend.update_tags(stream_id, tags)
 
@@ -342,6 +370,8 @@ class Datastream(object):
         :param tag: Dictionary describing the tag(s) to remove (values are ignored)
         """
 
+        self._check_tags(tag)
+
         self.backend.remove_tag(stream_id, tag)
 
     def clear_tags(self, stream_id):
@@ -364,6 +394,8 @@ class Datastream(object):
         :return: A `Streams` iterator over matched stream descriptors
         """
 
+        self._check_query_tags(query_tags or {})
+
         return self.backend.find_streams(query_tags)
 
     def append(self, stream_id, value, timestamp=None, check_timestamp=True):
@@ -384,7 +416,7 @@ class Datastream(object):
 
         return self.backend.append(stream_id, value, timestamp, check_timestamp)
 
-    def get_data(self, stream_id, granularity, start=None, end=None, start_exclusive=None, end_exclusive=None, reverse=False, value_downsamplers=None, time_downsamplers=None, limit=None):
+    def get_data(self, stream_id, granularity, start=None, end=None, start_exclusive=None, end_exclusive=None, reverse=False, value_downsamplers=None, time_downsamplers=None):
         """
         Retrieves data from a certain time range and of a certain granularity.
 
@@ -397,7 +429,6 @@ class Datastream(object):
         :param reverse: Should datapoints be returned in oldest to newest order (false), or in reverse (true)
         :param value_downsamplers: The list of downsamplers to limit datapoint values to (optional)
         :param time_downsamplers: The list of downsamplers to limit timestamp values to (optional)
-        :param limit: Limit the number of returned datapoints (optional)
         :return: A `Datapoints` iterator over datapoints
         """
 
@@ -434,13 +465,7 @@ class Datastream(object):
             if len(unsupported_downsamplers) > 0:
                 raise exceptions.UnsupportedDownsampler("Unsupported time downsampler(s): %s" % unsupported_downsamplers)
 
-        if limit is not None:
-            try:
-                limit = int(limit)
-            except ValueError:
-                raise ValueError("Limit value must be an integer!")
-
-        return self.backend.get_data(stream_id, granularity, start, end, start_exclusive, end_exclusive, reverse, value_downsamplers, time_downsamplers, limit)
+        return self.backend.get_data(stream_id, granularity, start, end, start_exclusive, end_exclusive, reverse, value_downsamplers, time_downsamplers)
 
     def downsample_streams(self, query_tags=None, until=None, return_datapoints=False):
         """
@@ -458,6 +483,8 @@ class Datastream(object):
                  for each datapoint created while downsampling, if `return_datapoints` was set
         """
 
+        self._check_query_tags(query_tags or {})
+
         if until is not None and until.tzinfo is None:
             until = until.replace(tzinfo=pytz.utc)
 
@@ -470,6 +497,8 @@ class Datastream(object):
         :param query_tags: Tags that should be matched to streams
         """
 
+        self._check_query_tags(query_tags or {})
+
         return self.backend.backprocess_streams(query_tags)
 
     def delete_streams(self, query_tags=None):
@@ -481,4 +510,6 @@ class Datastream(object):
         :param query_tags: Tags that should be matched to streams
         """
 
+        self._check_query_tags(query_tags or {})
+
         self.backend.delete_streams(query_tags)
