@@ -193,7 +193,7 @@ class ValueDownsamplers(DownsamplersBase):
         """
 
         name = 'count'
-        supports_types = ('numeric', 'graph')
+        supports_types = ('numeric', 'graph', 'nominal')
 
         def initialize(self):
             self.count = 0
@@ -924,6 +924,12 @@ class Streams(api.Streams):
         else:
             raise TypeError
 
+    def _get_backend_cursor(self):
+        if self.queryset is None:
+            return None
+
+        return self.queryset._cursor_obj
+
 
 class Datapoints(api.Datapoints):
     def __init__(self, datastream, cursor=None, empty_time=False):
@@ -959,6 +965,9 @@ class Datapoints(api.Datapoints):
         else:
             raise TypeError
 
+    def _get_backend_cursor(self):
+        return self.cursor
+
 
 class DownsampleState(mongoengine.EmbeddedDocument):
     timestamp = mongoengine.DateTimeField()
@@ -1051,9 +1060,9 @@ class Backend(object):
         """
 
         connection_settings.setdefault('tz_aware', True)
+        self.connection_settings = connection_settings
 
-        # Setup the database connection to MongoDB
-        mongoengine.connect(database_name, DATABASE_ALIAS, **connection_settings)
+        self._switch_database(database_name)
 
         assert \
             set(
@@ -1103,6 +1112,27 @@ class Backend(object):
         # Used only for concurrency tests
         self._test_concurrency = threading.local()
 
+    def _switch_database(self, database_name):
+        # Very internal. Just for debugging and testing.
+
+        # There can currently be only one connection to the database at a given moment, because
+        # stupid Mongoengine links all documents to the database alias. Even more, if a connection
+        # already exists with the same alias, then mongoengine.connect just returns the old connection,
+        # even if it is connected to a different database than requested! So to make sure, we first
+        # disconnect here, so that we get connected to the right database.
+        mongoengine.connection.disconnect(DATABASE_ALIAS)
+
+        # Settings are kept stored with old database name. Delete everything.
+        if DATABASE_ALIAS in mongoengine.connection._connection_settings:
+            del mongoengine.connection._connection_settings[DATABASE_ALIAS]
+
+        # And documents store collections with a link to old database. Clear that.
+        LastDatapoint._collection = None
+        Stream._collection = None
+
+        # Setup the database connection to MongoDB.
+        mongoengine.connect(database_name, DATABASE_ALIAS, **self.connection_settings)
+
     def ensure_stream(self, query_tags, tags, value_downsamplers, highest_granularity, derive_from, derive_op, derive_args, value_type):
         """
         Ensures that a specified stream exists.
@@ -1393,8 +1423,6 @@ class Backend(object):
 
         if prefix is None:
             prefix = 'tags'
-        else:
-            prefix = '%s__tags' % prefix
 
         for k, v in tags.iteritems():
             if isinstance(v, collections.Mapping):
@@ -1414,7 +1442,7 @@ class Backend(object):
         :param tags: A dictionary of new tags
         """
 
-        if not Stream.objects(external_id=uuid.UUID(stream_id)).update(**self._get_tag_query_dict('set', tags)):
+        if not Stream.objects(external_id=uuid.UUID(stream_id)).update(**self._get_tag_query_dict('set__tags', tags)):
             # Returned count is 1 if stream is found even if nothing changed, this is what we want
             raise exceptions.StreamNotFound
 
@@ -1426,7 +1454,7 @@ class Backend(object):
         :param tag: Dictionary describing the tag(s) to remove (values are ignored)
         """
 
-        if not Stream.objects(external_id=uuid.UUID(stream_id)).update(**self._get_tag_query_dict('unset', tag, 1)):
+        if not Stream.objects(external_id=uuid.UUID(stream_id)).update(**self._get_tag_query_dict('unset__tags', tag, 1)):
             # Returned count is 1 if stream is found even if nothing removed, this is what we want
             raise exceptions.StreamNotFound
 
@@ -1444,7 +1472,7 @@ class Backend(object):
             # Returned count is 1 if stream is found even if nothing changed, this is what we want
             raise exceptions.StreamNotFound
 
-    def _get_stream_queryset(self, query_tags):
+    def _get_stream_queryset(self, query_tags, allow_stream_id=True):
         """
         Returns a queryset that matches the specified stream tags.
 
@@ -1459,7 +1487,7 @@ class Backend(object):
 
         query_set = Stream.objects.all()
 
-        if 'stream_id' in query_tags:
+        if allow_stream_id and 'stream_id' in query_tags:
             query_set = query_set.filter(external_id=uuid.UUID(query_tags['stream_id']))
             del query_tags['stream_id']
 
@@ -1476,7 +1504,10 @@ class Backend(object):
         :return: A `Streams` iterator over matched stream descriptors
         """
 
-        return Streams(self, self._get_stream_queryset(query_tags))
+        # We do not specially process stream_id in find_streams. One should use get_tags
+        # instead. We force order by _id so that it is predictable and pagination works
+        # as expected. We do not provide API for custom sorting anyway.
+        return Streams(self, self._get_stream_queryset(query_tags, False).order_by('_id'))
 
     def _supported_timestamp_range(self, timestamp):
         """
@@ -1540,7 +1571,10 @@ class Backend(object):
             if isinstance(value, numbers.Number):
                 value = serialize_numeric_value(value)
             elif isinstance(value, dict):
-                # Assume that we are inserting a downsampled datapoint
+                # Assume that we are inserting a downsampled datapoint. Users can use this when the source
+                # of their values already provides information from multiple samples. For example, pinging
+                # over the Internet sends multiple packets and then returns min, max, mean times.
+                # By storing directly min, max, and mean values, no information is lost.
                 valid_keys = set()
                 for ds_name in stream.value_downsamplers:
                     key = api.VALUE_DOWNSAMPLERS[ds_name]
@@ -1557,6 +1591,12 @@ class Backend(object):
                 pass
             else:
                 raise TypeError("Streams of type 'numeric' may only accept numbers or downsampled datapoints!")
+        elif stream.value_type == 'nominal':
+            # We allow arbitrary values to be stored for nominal values. But do remember that they are stored
+            # as-is in the database so repeating the same huge value multiple times will be stored multiple
+            # times. If values will be repeating it is better to instead store only some small keys representing
+            # them.
+            pass
         elif stream.value_type == 'graph':
             if isinstance(value, dict):
                 try:
@@ -1716,7 +1756,7 @@ class Backend(object):
 
         return result
 
-    def get_data(self, stream_id, granularity, start=None, end=None, start_exclusive=None, end_exclusive=None, reverse=False, value_downsamplers=None, time_downsamplers=None, limit=None):
+    def get_data(self, stream_id, granularity, start=None, end=None, start_exclusive=None, end_exclusive=None, reverse=False, value_downsamplers=None, time_downsamplers=None):
         """
         Retrieves data from a certain time range and of a certain granularity.
 
@@ -1729,7 +1769,6 @@ class Backend(object):
         :param reverse: Should datapoints be returned in oldest to newest order (false), or in reverse (true)
         :param value_downsamplers: The list of downsamplers to limit datapoint values to (optional)
         :param time_downsamplers: The list of downsamplers to limit timestamp values to (optional)
-        :param limit: Limit the number of returned datapoints (optional)
         :return: A `Datapoints` iterator over datapoints
         """
 
@@ -1846,9 +1885,6 @@ class Backend(object):
             '_id': time_query,
         }, downsamplers).sort('_id', -1 if reverse else 1)
 
-        if limit is not None:
-            datapoints = datapoints.limit(limit)
-
         return Datapoints(self, datapoints, empty_time)
 
     def delete_streams(self, query_tags=None):
